# Contains all metric Protocols.

import datetime
from typing import Protocol


class Application(Protocol):
    """Base Protocol for an Application wrapper."""

    name: str


class ApplicationMetric(Protocol):
    """Base protocol for an Application Metric."""

    # The Application or Service to which the Metric Applies
    application: "Application"

    # This should be set to datetime.datetime.now() when compute() is called.
    computed_at: datetime.datetime | None

    # Stores errors that occured during the last computation.
    errors: list[Exception]

    @property
    def value(self) -> float | None:
        r"""Returns the computed metric value or None if not applicable."""
        ...

    def compute(self) -> bool:
        r"""A callable that (re)computes & stores the metric value.
        This method must handle all exceptions.

        Returns:
            bool: True if computation was successful, False otherwise.
        """
        ...



# Contains various metric runners

from typing import Enum, List, Union

from metrics import protocols
from metrics.types import OutputFormat


class BaseRunner:
    r"""Defines the base-runner for metrics generation."""


class MatrixRunner(BaseRunner):
    r"""A runner that builds & computes every metric for every application specified."""

    def run(
        self,
        metrics: protocols.ApplicationMetric | List[protocols.ApplicationMetric],
        applications: Union[
            List[protocols.Application], List[str], protocols.Application, str
        ],
        format: Enum = OutputFormat.LOG,
    ) -> List[protocols.ApplicationMetric]:
        r"""Builds & Runs metric generation."""

        if not isinstance(metrics, list):
            metrics = [metrics]

        if not isinstance(applications, list):
            applications = [applications]

        if not isinstance(format, OutputFormat):
            raise TypeError(
                f"Expected format type to be an {type(OutputFormat)}, got {type(format)}"
            )

        application_metrics = []
        for application in applications:
            # If string passed, convert to Application object.
            if isinstance(application, str):
                application = protocols.Application(name=application)

            for metric in metrics:
                metric = metric(application)
                metric.compute()
                application_metrics.append(metric)

                if format == OutputFormat.OUTPUT_FORMAT_LOG:
                    print(metric)
                    print(f"{metric.application.name:<20}: {metric.value:.1f}")

        return application_metrics



from enum import Enum, auto


class OutputFormat(Enum):
    """Defines the output formats for metrics generation."""

    OUTPUT_FORMAT_UNSPECIFIED = auto()
    OUTPUT_FORMAT_LOG = auto()
    OUTPUT_FORMAT_MARKDOWN = auto()
    OUTPUT_FORMAT_JSON = auto()
    OUTPUT_FORMAT_YAML = auto()



import datetime
from abc import ABC, abstractmethod
from typing import Any, Final, List, Optional


class Application(ABC):
    """Base class for an Application wrapper."""

    @abstractmethod
    def __init__(self, name: str):
        self.name = name


class ApplicationMetric(ABC):
    """Defines an Application specific Metric"""

    @abstractmethod
    def __init__(self, application: Application):
        """Initialies the metric with the application it is associated to."""
        self.application: Final[Application] = application
        self.computed_at: Optional[datetime.datetime] = None
        self.errors: List[Exception] = []
        self._value: Optional[Any] = None

    @property
    def value(self) -> Any:
        """Fixed value of the metric."""
        return self._value

    @value.setter
    def value(self, value: Any) -> None:
        r"""This set's the metric value and updates the computed_at timestamp."""
        self.computed_at = datetime.datetime.now()
        self._value = value

    @abstractmethod
    def compute(self) -> (bool, Any):
        ...

    def asdict(self) -> dict[str, Any]:
        return {
            "application": self.application.name,
            "metric": self.__class__.__name__,
            "computed_at": self.computed_at,
            "errors": self.errors,
            "value": self.value,
        }


class NumericApplicationMetric(ApplicationMetric):
    @property
    @abstractmethod
    def value(self) -> Optional[float]:
        return self._value

    def compute(self) -> (bool, Optional[float]):
        ...
